/*  
 Copyright (C) 2010 - 2011  Quantum Nomad LLC (http://www.qnomad.com/)

 This program is licensed under version 3 of the GNU General Public License as published
 by the Free Software Foundation.  You should have received a copy of the license along
 with the source code.  If not, an online copy of the license can be found at
 http://www.qnomad.com/wpf/GPL.aspx.
*/

using System;
using System.Collections;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.Reflection;

namespace Qnomad.BidirectionalAssociations {
	/// <summary>
	/// Keeps both sides of a bidirectional many-to-many association in sync with each other.
	/// </summary>
	/// <author>Adrian Alexander</author>
	public sealed class ManyToManyAssocSync<T> {
		private T thisSide;
		private string otherSidePropertyName;
		private PropertyInfo otherSideProperty;

		public ManyToManyAssocSync( T thisSide, string otherSideToThisSidePropertyName ) {
			this.thisSide = thisSide;
			this.otherSidePropertyName = otherSideToThisSidePropertyName;
		}

		/// <summary>
		/// Responds to add/remove events raised by this side's collection.
		/// </summary>
		public void UpdateOtherSide( object sender, NotifyCollectionChangedEventArgs e ) {
			if ( e.Action == NotifyCollectionChangedAction.Add )
				//addingToOtherSide: the item that was just added to this side's collection
				foreach ( object addingToOtherSide in e.NewItems ) {
					try {
						ICollection<T> otherSidesCollection = GetOtherSidesCollection( addingToOtherSide );
						if ( ReflectionExtentions.IsInitialized(otherSidesCollection) && !otherSidesCollection.Contains(thisSide) ) {
							if ( otherSidesCollection is IList && otherSidesCollection.Contains(thisSide) ) // true if add will cause duplicates
								throw new InvalidOperationException( "Collection is non-unique" );
							otherSidesCollection.Add( thisSide );
						}
					} catch ( TargetException ) { }
				}
			else if ( e.Action == NotifyCollectionChangedAction.Remove )
				//removingFromOtherSide: the item that was just removed from this side's collection
				foreach ( object removingFromOtherSide in e.OldItems ) {
					try {
						ICollection<T> otherSidesCollection = GetOtherSidesCollection( removingFromOtherSide );
						if ( ReflectionExtentions.IsInitialized(otherSidesCollection) && otherSidesCollection.Contains(thisSide) ) {
							otherSidesCollection.Remove( thisSide );
							if ( otherSidesCollection is IList && otherSidesCollection.Contains(thisSide) ) // true if there was more than one
								throw new InvalidOperationException( "Collection is non-unique" );
						}
					} catch ( TargetException ) { }
				}
		}

		private ICollection<T> GetOtherSidesCollection( object otherSide ) {
			if ( otherSideProperty == null )
				otherSideProperty = otherSide.GetType().GetProperty( otherSidePropertyName );
			return (ICollection<T>)otherSideProperty.GetValue( otherSide, null );
		}
	}
}
