﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Diagnostics;

namespace Qnomad
{
	public abstract class ChangeNotifier : INotifyPropertyChanged
	{
		#region PropertyChanged Event

		public event PropertyChangedEventHandler PropertyChanged;

		protected bool SetProperty<T>( ref T backingField, T value, string propertyName )
		{
			bool equal = EqualityComparer<T>.Default.Equals( backingField, value );
			return SetProperty( !equal, ref backingField, value, propertyName );
		}

		protected bool SetProperty( ref Uri backingField, Uri value, string propertyName )
		{
			bool equal = ReferenceEquals( backingField, value ) ||
				(backingField != null && value != null && backingField.ToString() == value.ToString());
			return SetProperty( !equal, ref backingField, value, propertyName );
		}

		private bool SetProperty<T>( bool change, ref T backingField, T value, string propertyName )
		{
			if ( change ) {
				backingField = value;
				VerifyPropertyName( propertyName );
				OnPropertyChanged( propertyName );
			}
			return change;
		}

		protected virtual void OnPropertyChanged( string propertyName )
		{
			if ( PropertyChanged != null )
				PropertyChanged( this, new PropertyChangedEventArgs( propertyName ) );
		}

		#endregion PropertyChanged Event

		#region Debugging Aides

		/// <summary>
		/// Verifies that the specified property name matches the actual property name.
		/// This method does not exist in a Release build.
		/// </summary>
		[Conditional( "DEBUG" )]
		[DebuggerStepThrough]
		private void VerifyPropertyName( string propertyName )
		{
			string actualPropertyName = GetPropertyNameFromStackTrace();

			// Verify that propertyName matches the name found in the stack trace.
			if ( !propertyName.Equals( actualPropertyName ) ) {
				string mesg = "An incorrect property name of '{0}' was specified for the '{1}' property.";
				mesg = string.Format( mesg, propertyName, actualPropertyName );
				throw new InvalidOperationException( mesg );
			}
		}

		[DebuggerStepThrough]
		private static string GetPropertyNameFromStackTrace()
		{
			var stackTrace = new StackTrace(); //get call stack
			var stackFrames = stackTrace.GetFrames(); //get method calls (frames)

			foreach ( StackFrame f in stackFrames ) {
				string name = f.GetMethod().Name;

				if ( name.StartsWith( "set_", StringComparison.Ordinal ) )
					return name.Substring( 4 );
			}

			string s = string.Empty;
			foreach ( StackFrame f in stackFrames ) {
				s += "   " + f.GetMethod().Name + Environment.NewLine;
			}
			throw new Exception( "Could not determine property name. StackTrace:\n" + s );
		}

		#endregion Debugging Aides

		/*Message message;
		public Message Message
		{
			get { return message; }
			set { SetProperty( ref message, value, "Message" ); }
		}*/
	}
}
