﻿/*  
 Copyright (C) 2010 - 2011  Quantum Nomad LLC (http://www.qnomad.com/)

 This program is licensed under version 3 of the GNU General Public License as published
 by the Free Software Foundation.  You should have received a copy of the license along
 with the source code.  If not, an online copy of the license can be found at
 http://www.qnomad.com/wpf/GPL.aspx.
*/

using System;
using System.Collections.Generic;
using System.Linq;
using System.Windows.Input;

namespace Qnomad.Outlining
{
	/// <summary>
	/// Encapsulates an operation to place a node within a tree and, if already present in the tree,
	/// remove it from its old location. The new position can be described in one of two ways:
	///   (a) as a "nudge" relative to the original position of the node being moved (frontward, rearward, demote, promote), or
	///   (b) as an "insert" relative to another node that acts as a target (sibling after, sibling before, child, parent).
	/// </summary>
	/// <typeparam name="T"></typeparam>
	public class HierarchicalPlacementArgs<T> : FlatPlacementArgs<T, T>
	  where T : ITreeNode<T>
	{
		/// <summary>Siblings of the new item, that will become its children</summary>
		public IList<T> itemsToSubsume;

		public HierarchicalPlacementArgs( RelativeInsert relativeInsert, T itemToMove, T targetLoc, bool isSorted )
			: base( relativeInsert, itemToMove, targetLoc, isSorted ) { }

		public HierarchicalPlacementArgs( ICommand command, T target, bool isSorted )
			: base( command, target, isSorted ) { }

		public HierarchicalPlacementArgs( ICommand command, IList<T> targets, bool isSorted )
			: base( command, targets, isSorted ) { }

		protected override bool ProcessWithin( T itemToMove, T targetLoc )
		{
			newPosition.parent = targetLoc;
			return !newPosition.parent.AncestorsAndSelf().Contains( itemToMove ); //cannot drop an item into itself
		}

		protected override bool SetNudgePosition( ICommand command, T target )
		{
			bool result = base.SetNudgePosition( command, target );
			if ( !result ) return false; //finished if already invalid

			if ( command == OutliningCommands.NudgeDemote ) {
				if ( oldPosition.childIndex == 0 ) return false; //Nowhere to demote to!
				newPosition.parent = ItemToMove.Parent.SubItems[oldPosition.childIndex - 1];
			} else if ( command == OutliningCommands.NudgePromote ) {
				var oldGrandParent = oldPosition.parent.Parent;
				if ( oldGrandParent == null ) return false; //Already at the top level!
				newPosition.parent = oldGrandParent;
				newPosition.childIndex = oldGrandParent.SubItems.IndexOf( oldPosition.parent ) + 1;
			}
			SetSortedIndex();
			return result;
		}

		protected override bool SetInsertPosition( ICommand command, IList<T> targets )
		{
			bool result = base.SetInsertPosition( command, targets );
			if ( !result ) return false; //finished if already invalid

			if ( command == OutliningCommands.InsertChild )
				newPosition.parent = targets[0];
			else if ( command == OutliningCommands.InsertParent ) {
				newPosition.parent = targets[0].Parent;
				if ( newPosition.parent == null ) return false;
				newPosition.childIndex = MinimumIndex( newPosition.parent.SubItems, (this.itemsToSubsume = targets) );
			}
			return result;
		}

		public override void Apply()
		{
			// Insert first so that below's Add() call will reuse item's previous HierarchicalPresentationModel.
			base.Apply();

			IList<T> subItems = this.newPosition.parent.SubItems;
			if ( this.itemsToSubsume != null ) { //If originated from InsertParent command:
				foreach ( T item in this.itemsToSubsume ) { //loop thru selected items:
					if ( !subItems.Contains(item) )
						throw new Exception( "Parent's subitems must contain new item's children" );
					subItems.Remove( item ); //remove item from parent's sub-items list
					ItemToMove.SubItems.Add( item ); //add it as child of the new item
				}
			}
		}

		public override string ToString()
		{
			string result = base.ToString();
			if ( itemsToSubsume != null ) {
				result += ", itemsToSubsume=";
				foreach ( T i in itemsToSubsume )
					result = result + "\n\t" + i;
			}
			return result;
		}
	}
}
