﻿using System;
using System.Collections.Generic;
using System.Linq;

namespace Qnomad.Outlining
{
	internal static class TreeOperations
	{
		#region 'Descendants' Iterators

		/// <summary>
		/// Returns an iterator that performs a breadth-first traversal of nodes.
		/// </summary>
		internal static IEnumerable<T> DescendantsAndSelf<T>( T root, ITreeNodeAdapter<T> adapter )
		{
			var queue = new Queue<T>();
			queue.Enqueue( root );

			while ( queue.Count > 0 ) {
				T head = queue.Dequeue();
				yield return head;

				if ( adapter.HasSubItems(head) ) {
					foreach ( T child in adapter.GetSubItems(head) )
						queue.Enqueue( child );
				}
			}
		}

		internal static IEnumerable<T> Descendants<T>( T root, ITreeNodeAdapter<T> adapter )
		{
			return DescendantsAndSelf( root, adapter ).Skip( 1 );
		}

		#endregion Descendants

		#region 'Ancestors' Iterators

		internal static IEnumerable<T> Ancestors<T>( T startNode, ITreeNodeAdapter<T> adapter )
		{
			var parent = adapter.GetParent( startNode );
			while ( parent != null ) {
				yield return parent;
				parent = adapter.GetParent( parent );
			}
		}

		internal static IEnumerable<T> AncestorsAndSelf<T>( T startNode, ITreeNodeAdapter<T> adapter )
		{
			yield return startNode;

			foreach ( var ancestor in Ancestors(startNode, adapter) )
				yield return ancestor;
		}

		#endregion Ancestors

		#region 'Siblings' Iterators

		internal static IEnumerable<T> PrecedingSiblings<T>( T context, ITreeNodeAdapter<T> adapter )
		{
			var parent = adapter.GetParent( context );
			if ( parent == null ) yield break;

			foreach ( T sibling in adapter.GetSubItems(parent) ) {
				if ( sibling.Equals(context) ) break;
				yield return sibling;
			}
		}

		internal static IEnumerable<T> FollowingSiblings<T>( T context, ITreeNodeAdapter<T> adapter )
		{
			var parent = adapter.GetParent( context );
			if ( parent == null ) yield break;

			bool afterSelf = false;
			foreach ( T sibling in adapter.GetSubItems(parent) ) {
				if ( afterSelf )
					yield return sibling;

				if ( sibling.Equals(context) )
					afterSelf = true;
			}
		}

		#endregion Siblings
	}
}
